/* $Id$ */

/*========================================================================
 *  Copyright (c) Michael J. Hammel 1995,1996,1997,1998.
 *========================================================================
 *              FILE NAME: pilotaddr.c
 *            DESCRIPTION: Parser routines for pilot address files
 *      DEFINED CONSTANTS: 
 *       TYPE DEFINITIONS: 
 *      MACRO DEFINITIONS: 
 *       GLOBAL VARIABLES: 
 *       PUBLIC FUNCTIONS: XPPilotParseAddr
 *      PRIVATE FUNCTIONS: ProcessFields
 *  SOFTWARE DEPENDENCIES: 
 *  HARDWARE DEPENDENCIES: 
 *                  NOTES: 
 *
 * SPECIAL CONSIDERATIONS:
 * Set your tabstops to 3 to make the code more readable.
 *========================================================================
 *
 * MODIFICATION HISTORY:
 * $Log$
 *
 *========================================================================*/
#define XP_CPADDR_C

/* === System Headers === */
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/stat.h>
#include <sys/types.h>

#ifdef DO_GTK
#include <gtk/gtk.h> 
#else
#include "motif.h"
#endif

/* === Project Headers === */
#include "XNotesPlus.h"
#include "debug.h"

/* === external routines === */

/* === Public routine prototypes === */

/* === Private routine prototypes === */
static void ProcessFields();
static void PrintFields();

/* === Global Variables === */
PilotAddr_T			*Pilot_Address=NULL;


/*========================================================================
 *	Name:			XPPilotParseAddr
 *	Prototype:	void XPPilotParseAddr(char *filename)
 *					
 *	Description:
 *	Parses the CSV output from pilot-addresses
 *
 *	Input Arguments:
 *		char	*filename		name of file with addresses in it.
 *	Output Arguments:
 *	Return Values:
 *		XP_E_FILE_EXIST		file to parse doesn't exist or is not a text file
 *	Global Variables:
 *	External Routines:
 *	Method:
 *	Restrictions:
 *	Notes:
 * I couldn't find any details on the CSV format, so I'm just using the
 * output from pilot-addresses to determine how to parse the input.
 *		
 *========================================================================*/
int 
XPPilotParseAddr(
	char	*filename
)
{
#ifdef DEBUG
	char				*fname="XPPiloParseAddr()";
#endif

#ifdef DO_MOTIF
	Boolean			in_field, escape;
#else
	gboolean			in_field, escape;
#endif

	struct stat		stat_buf;
	char				buf[4096];
	FILE				*fd;
	char				c;
	int				field_count;
	char				*ptr;
	PilotAddr_T		*next_link;
	PilotAddr_T		*current_link;

	DBGEnter();
	DBGPrintf(DBG_INFO,("pilot db to look for: %s\n", filename));

	/*
	 * Verify file exists.  Return if it doesn't.
	 */
	if ( stat(filename, &stat_buf) != 0 )
		return XP_E_FILE_EXIST;

	if ( ! ( stat_buf.st_mode & S_IFREG ) )
		return XP_E_FILE_EXIST;

	/*
	 * Free up any old link lists.
	 */
	current_link = Pilot_Address;
	while (current_link != NULL )
	{
		next_link = current_link->next;
		free(current_link);
		current_link = next_link;
	}
	Pilot_Address=NULL;

	/*
	 * File is ok.  Open it.
	 */
	fd = fopen(filename, "r");

	/*
	 * Here is the ugly part.  Read in the file, one character at a time.
	 * Bleck.  We have to maintain state so we know:
	 * 1. are we in a record or moving to a new record?
	 * 2. are we in a field or at a field delimiter?
	 */

	in_field = False;
	escape = False;
	field_count = 0;
	ptr=buf;

	while ( (c=fgetc(fd)) != EOF ) 
	{

		/*
		 * If we're in a field, then we either copy the character to a 
		 * buffer, or if we've found the end of the field we process the
		 * buffer and prep for a new field.
		 */
		if ( in_field )
		{
			switch (c)
			{

				/*
				 * Skip escape marker - but flag that the next character is
				 * supposed to be part of the field and not a delimiter.
				 */
				case '\\':	
					escape=True;
					break;

				/*
				 * If the character is a double quote and it has not been
				 * escaped then we're at the end of a field.  If it has
				 * been escaped, copy to the field buffer.
				 */
				case '\"':	
					if ( escape )
					{
						bcopy((char *)&c, ptr++, 1);
						escape=False;
					}
					else
					{
						bcopy("\0", ptr++, 1);
						in_field = False; 
						if (field_count > 20 )
							field_count = 1;
						ProcessFields(field_count, buf);
						ptr=buf;
					}
					break;

				/*
				 * Everything else is copied to the field buffer.
				 */
				default: 
					bcopy((char *)&c, ptr++, 1);
					break;
			}
		}
		else
		{
			/*
			 * Not in a field yet - look for start of field marker (a 
			 * double quote).
			 */
			switch (c)
			{

				/*
				 * The double quote marks the start of a new field.
				 */
				case '\"':	
					in_field = True; 
					field_count++;
					break;

				/*
				 * I think that Private (or category) can be just plain
				 * 0 (no double quotes).  Here is where I'd handle that.
				 */


				/*
				 * Toss out anything else (probably just commas).
				 */
				default:
					break;
			}
		}
	}

#ifdef DEBUG
	PrintFields();
#endif
	return ( XP_E_SUCCESS );
}


/*========================================================================
 *	Name:			ProcessFields
 *	Prototype:	void ProcessFields(int count, char *field)
 *					
 *	Description:
 *	Process each field as it is passed in by adding it to the appropriate
 * field of an allocated record.  Allocate records as needed.
 *
 *	Input Arguments:
 *		int	count				field identifier
 *		char	*field		field data
 *========================================================================*/
static void
ProcessFields(
	int	count,
	char	*field
)
{
#ifdef DEBUG
	char				*fname="ProcessFields()";
#endif


	PilotAddr_T					*pentry;
	static PilotAddr_T		*current_link=NULL;

/*
	DBGEnter();
*/

	switch(count)
	{
		case 1: 
			/*
			 * New record.  Allocate a new data structure for it, then
			 * initialize the record.
			 */
			DBGPrintf(DBG_INFO,("Start new record: %s\n", field));
			pentry = (PilotAddr_T *)XPSafeAlloc(sizeof(PilotAddr_T));
			pentry->last = NULL;
			pentry->first = NULL;
			pentry->title = NULL;
			pentry->company = NULL;
			pentry->work = NULL;
			pentry->home = NULL;
			pentry->fax = NULL;
			pentry->other = NULL;
			pentry->email = NULL;
			pentry->address = NULL;
			pentry->city = NULL;
			pentry->state = NULL;
			pentry->zip = NULL;
			pentry->country = NULL;
			pentry->custom1 = NULL;
			pentry->custom2 = NULL;
			pentry->custom3 = NULL;
			pentry->custom4 = NULL;
			pentry->note = NULL;
			pentry->private = NULL;
			pentry->category = NULL;
				
			if ( Pilot_Address == NULL )
				Pilot_Address = pentry;
			else
				current_link->next = pentry;

			current_link = pentry;
			current_link->next = NULL;

			/* Fill in the last name field */
			current_link->last = XPSafeAlloc(strlen(field) + 1);
			strcpy(current_link->last, field);
			break;

		/*
		 * Fill in the appropriate field of the current record
		 * with the field passed to us.
		 */
		case 2: 
			current_link->first = XPSafeAlloc(strlen(field) + 1);
			strcpy(current_link->first, field);
			break;

		case 3: 
			current_link->title = XPSafeAlloc(strlen(field) + 1);
			strcpy(current_link->title, field);
			break;

		case 4: 
			current_link->company = XPSafeAlloc(strlen(field) + 1);
			strcpy(current_link->company, field);
			break;

		case 5:
			current_link->work = XPSafeAlloc(strlen(field) + 1);
			strcpy(current_link->work, field);
			break;

		case 6: 
			current_link->home = XPSafeAlloc(strlen(field) + 1);
			strcpy(current_link->home, field);
			break;

		case 7: 
			current_link->fax = XPSafeAlloc(strlen(field) + 1);
			strcpy(current_link->fax, field);
			break;

		case 8: 
			current_link->other = XPSafeAlloc(strlen(field) + 1);
			strcpy(current_link->other, field);
			break;

		case 9: 
			current_link->email = XPSafeAlloc(strlen(field) + 1);
			strcpy(current_link->email, field);
			break;

		case 10: 
			current_link->address = XPSafeAlloc(strlen(field) + 1);
			strcpy(current_link->address, field);
			break;

		case 11: 
			current_link->city = XPSafeAlloc(strlen(field) + 1);
			strcpy(current_link->city, field);
			break;

		case 12: 
			current_link->state = XPSafeAlloc(strlen(field) + 1);
			strcpy(current_link->state, field);
			break;

		case 13: 
			current_link->zip = XPSafeAlloc(strlen(field) + 1);
			strcpy(current_link->zip, field);
			break;

		case 14: 
			current_link->country = XPSafeAlloc(strlen(field) + 1);
			strcpy(current_link->country, field);
			break;

		case 15: 
			current_link->custom1 = XPSafeAlloc(strlen(field) + 1);
			strcpy(current_link->custom1, field);
			break;

		case 16: 
			current_link->custom2 = XPSafeAlloc(strlen(field) + 1);
			strcpy(current_link->custom2, field);
			break;

		case 17: 
			current_link->custom3 = XPSafeAlloc(strlen(field) + 1);
			strcpy(current_link->custom3, field);
			break;

		case 18: 
			current_link->custom4 = XPSafeAlloc(strlen(field) + 1);
			strcpy(current_link->custom4, field);
			break;

		case 19: 
			current_link->note = XPSafeAlloc(strlen(field) + 1);
			strcpy(current_link->note, field);
			break;

		case 20: 
			current_link->private = XPSafeAlloc(strlen(field) + 1);
			strcpy(current_link->private, field);
			break;

	}

}

static void
PrintFields()
{
#ifdef DEBUG
	char		*fname="PrintFields()";
#endif

	PilotAddr_T		*current_link;
	
	DBGEnter();

	current_link = Pilot_Address;
	DBGPrintf(DBG_TRACE3, ("New record: \n"));
	while (current_link != NULL )
	{
		DBGWrite(DBG_TRACE3,
			("\tlast: %s\n", current_link->last));

		current_link = current_link->next;
	}
}
