/* $Id$ */

/*========================================================================
 *  Copyright (c) Michael J. Hammel 1995,1996,1997,1998.
 *========================================================================
 *              FILE NAME: config.c
 *            DESCRIPTION: config file routines for XNotesPlus notes
 *      DEFINED CONSTANTS: 
 *       TYPE DEFINITIONS: 
 *      MACRO DEFINITIONS: 
 *       GLOBAL VARIABLES: 
 *       PUBLIC FUNCTIONS: XPReadConfigFile, ALConvertCase
 *      PRIVATE FUNCTIONS: utilGetWordA, UpdateAppRes
 *  SOFTWARE DEPENDENCIES: 
 *  HARDWARE DEPENDENCIES: 
 *                  NOTES: 
 *	
 *
 * SPECIAL CONSIDERATIONS:
 * Set your tabstops to 3 to make the code more readable.
 *
 *========================================================================
 *
 * MODIFICATION HISTORY:
 * $Log$
 *
 *========================================================================*/
#define XP_CONFIG_C

/* === System Headers === */
#include <X11/Xlib.h>
#include <string.h>
#include <time.h>
#include <sys/param.h>
#ifdef USE_DIR
#include <sys/dir.h>
#else
#include <dirent.h>
#endif
#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <ctype.h>

/* GTK/GDK specific includes */
#include <gtk/gtk.h>
#include <gdk/gdkx.h>


/* === Project Headers === */
#include "XNotesPlus.h"
#include "debug.h"


/* === external routines === */


/* === Public routine prototypes === */
void ALConvertCase();


/* === Private routine prototypes === */
static void	UpdateAppRes();
static char *utilGetWordA(char *buf, char delimiter);


/* === Global Variables === */
extern AppRes		app_res;
extern StickyNote	*notes;
extern GtkWidget	*toplevel;


/* === Static Variables === */



/*========================================================================
 *	Name     : XPReadConfigFile
 *	Prototype: int XPReadConfigFile()
 *					
 *	Description:
 *		Read in the config file and parse options.
 *		Config file lines have the format
 *			keyword:value
 *		Comments are lines which begin with ##
 *		Blank lines and leading white space is ignored.
 *		Configuration lines must not contain newline (\n) characters.
 *
 *	Input Arguments:
 *		config_file			name of configuration file to read
 *
 *	Return Values:
 *		XP_E_FAILURE		config file error
 *		
 *========================================================================*/
int
XPReadConfigFile()
{
	struct stat		stat_buf;

	char	line[4098];
	char	utilError[256];
	char	*chr, *name, *val;
	FILE	*fd;
	char	*config_file;
	

	/*
	 * If the XPCONFIGFILE environment variable is set, use that.
	 * Otherwise, assume default notes directory has the config file.
	 */
	if ( (config_file=getenv("XPCONFIGFILE")) == NULL )
	{
		app_res.config_file = 
			XPSafeAlloc(strlen(app_res.home_dir) + 
				strlen(app_res.note_dir) + strlen(DefaultConfigFile) + 2 + 1);
		strcpy(app_res.config_file, (char *)app_res.home_dir);
		strcat(app_res.config_file, (char *)"/");
		strcat(app_res.config_file, (char *)app_res.note_dir);
		strcat(app_res.config_file, (char *)"/");
		strcat(app_res.config_file, DefaultConfigFile);
		config_file = app_res.config_file;
	}
	else
	{
		app_res.config_file = XPSafeAlloc(strlen(config_file) + 1);
		strcpy(app_res.config_file, (char *)config_file);
	}


	/*
	 * Make sure config file exists
	 */
	if ( stat( config_file, &stat_buf ) == -1 )
	{
		sprintf ( utilError, XP_S_NO_CONFIG_FILE, XP_S_PROG_NAME, config_file );
		fprintf(stderr, utilError);
		return ( XP_E_FAILURE );
	}

	/* 
	 * read in the config information
	 */
	fd = fopen ( config_file, "r" );
		
	while ( fgets( line, 4098, fd ) != NULL )
	{
		/*
		 * If the line is not a comment...
		 */
		if ( strncmp( line, "#", 1 ) != 0 )
		{
			/*
			 * Move past white space
			 */
			chr=line;
			while (	(isspace ((int)*chr) != 0 ) && 
						(chr != (line+strlen(line)) ) )
				chr++;

			if (chr != (line+strlen(line)) )
			{

				/*
				 * remove the newline
				 */
				strncpy ( (char *)(line+strlen(line)-1), "\0", 1 );
	
				/*
				 * Pull the keyword and value from the line
				 */
				name = utilGetWordA(chr, XP_DELIMITER);
				
				if ( strlen ( chr ) > 0 )
				{
					val= (char *)XPSafeAlloc(strlen(chr)+1);
					strcpy(val, chr);
				}
				else
					val=NULL;
		
				/*
				 * Update the application resources structure.
			 	 */
				UpdateAppRes(name, val);
		
				/*
				 * Free up storage.
				 */
				free(name);
				free(val);
			}
		}
	}

	return(0);
}

/*========================================================================
 *	Name     : utilGetWordA			 
 *	Prototype: char *utilGetWordA(char *line, char stop)
 *
 *	Description:
 * Parse a word, delimited by the stop character, from line.  Update
 * line to point to character immediately after stop character.
 *	Allocates the storage for the parsed word.
 *
 *	Input Arguments:
 *		char *line				line to parse word from
 *		char stop				character to use as the delimiter between words
 *
 *	Return Values:
 *		char *word				points to word parsed by utilGetWordA()
 *
 *	Restrictions:
 *		Caller is responsible for cleaning up storage allocated by this
 *		routine.
 *		
 *========================================================================*/
static char *
utilGetWordA(
	char *line, 
	char stop
)
{
    int x = 0,y;
    char *word = (char *) XPSafeAlloc(sizeof(char) * (strlen(line) + 1));

    for(x=0;((line[x]) && (line[x] != stop));x++)
        word[x] = line[x];

    word[x] = '\0';
    if(line[x]) ++x;
    y=0;

    while( (line[y++] = line[x++]) );
    return word;
}



/*========================================================================
 *	Name:			ALConvertCase
 *	Prototype:	ALConvertCase(char *buf, AL_CASE_TYPE type, int requested_length)
 *
 *	Description:
 *		Convert a string to upper or lower case.
 *
 *	Input Arguments:
 *		char				*buf						buffer to convert
 *		AL_CASE_TYPE	type						AL_UPPER or AL_LOWER
 *		int				requested_length		If not 0 then use this value for 
 *														number of characters to convert.
 *
 *	Output Arguments:
 *		buf is converted in place
 *
 *	Return Values:
 *	Global Variables:
 *	External Routines:
 *	Method:
 *	Restrictions:
 *	Notes:
 *		
 *========================================================================*/
void
ALConvertCase ( 
	char				*buf,
	AL_CASE_TYPE	type,
	int				requested_length
)
{
	int	i;
	char	ch1,ch2;
	int	length;


	if (requested_length == 0)
		length = strlen( buf );
	else
		length = requested_length;

	for ( i = 0; i < length; i++ )
	{
		strncpy ( (char *)&ch1, buf,1 );
		if ( type == AL_LOWER )
			ch2 = tolower ( ch1 ); 
		else
			ch2 = toupper ( ch1 ); 
		strncpy ( buf, (char *)&ch2, 1 );
		buf++;
	}
}


/*========================================================================
 *	Name     : UpdateAppRes
 *	Prototype: void UpdateAppRes(char *name, char *val)
 *
 *	Description:
 * Stuff the value specified into the appropriate structure member
 * of the application resources data structure.
 *
 *	Input Arguments:
 *		char *name				config file token
 *		char *val				value parsed for this token
 *
 *	Restrictions:
 *	If the name token is unknown, its is ignored.
 *
 * Notes:
 * There must be an easier way to do this.  Certainly there are more 
 * sophisticated ways of doing it.  But what the hell.  This should work.
 *========================================================================*/
static void
UpdateAppRes(
	char *name, 
	char *val
)
{
	/*
	 * Convert the two strings to uppercase.
	 */
	ALConvertCase(name, AL_UPPER, 0);

	/*
	 * Home directory configuration.
	 */
	if ( strcmp(name, XP_CFG_HOME) == 0 )
	{
		if ( app_res.home_dir )
			free( app_res.home_dir );
		app_res.home_dir = XPSafeAlloc(strlen(val)+1);
		strcpy(app_res.home_dir, (char *)val);
		return;
	}

	/*
	 * Note directory configuration.
	 */
	if ( strcmp(name, XP_CFG_NOTEDIR) == 0 )
	{
		if ( app_res.note_dir )
			free(app_res.note_dir);
		app_res.note_dir = XPSafeAlloc(strlen(val)+1);
		strcpy(app_res.note_dir, (char *)val);
		return;
	}

	/*
	 * Temporary directory configuration.
	 */
	if ( strcmp(name, XP_CFG_TMPDIR) == 0 )
	{
		if ( app_res.tmp_dir )
			free(app_res.tmp_dir);
		app_res.tmp_dir = XPSafeAlloc(strlen(val)+1);
		strcpy(app_res.tmp_dir, (char *)val);
		return;
	}

	/*
	 * XPM directory configuration.
	 */
	if ( strcmp(name, XP_CFG_XPMDIR) == 0 )
	{
		if ( app_res.xpm_dir )
			free(app_res.xpm_dir);
		app_res.xpm_dir = XPSafeAlloc(strlen(val)+1);
		strcpy(app_res.xpm_dir, (char *)val);
		return;
	}

	/*
	 * Audio files directory.
	 */
	if ( strcmp(name, XP_CFG_AUDIODIR) == 0 )
	{
		if ( app_res.audio_dir )
			free(app_res.audio_dir);
		app_res.audio_dir = XPSafeAlloc(strlen(val)+1);
		strcpy(app_res.audio_dir, (char *)val);
		return;
	}

	/*
	 * Pilot Directory configuration.
	 */
	if ( strcmp(name, XP_CFG_PILOTDIR) == 0 )
	{
		if ( app_res.pilot_dir )
			free( app_res.pilot_dir );
		app_res.pilot_dir = XPSafeAlloc(strlen(val)+1);
		strcpy(app_res.pilot_dir, (char *)val);
		return;
	}

	/*
	 * Address directory configuration.
	 */
	if ( strcmp(name, XP_CFG_ADDRDIR) == 0 )
	{
		if ( app_res.addr_dir )
			free( app_res.addr_dir );
		app_res.addr_dir = XPSafeAlloc(strlen(val)+1);
		strcpy(app_res.addr_dir, (char *)val);
		return;
	}

	/*
	 * NoAudio configuration.
	 */
	if ( strcmp(name, XP_CFG_NOAUDIO) == 0 )
	{
		ALConvertCase(val, AL_UPPER, 0);
		if ( strcmp(val, XP_CFG_YES) != 0 )
			app_res.noaudio = FALSE;
		else
			app_res.noaudio = TRUE;
		return;
	}

	/*
	 * Autosave interval configuration.
	 */
	if ( strcmp(name, XP_CFG_AUTOSAVE_INTERVAL) == 0 )
	{
		app_res.interval = atoi(val);
		return;
	}

	/*
	 * Save on Exit configuration.
	 */
	if ( strcmp(name, XP_CFG_SAVEEXIT) == 0 )
	{
		ALConvertCase(val, AL_UPPER, 0);
		if ( strcmp(val, XP_CFG_YES) == 0 )
			app_res.save_notes = TRUE;
		else
			app_res.save_notes = FALSE;
		return;
	}

	/*
	 * Autosave configuration.
	 */
	if ( strcmp(name, XP_CFG_AUTOSAVE) == 0 )
	{
		ALConvertCase(val, AL_UPPER, 0);
		if ( strcmp(val, XP_CFG_ON) != 0 )
			app_res.nosave = FALSE;
		else
			app_res.nosave = TRUE;
		return;
	}

	/*
	 * Alarm configuration.
	 */
	if ( strcmp(name, XP_CFG_ALARM) == 0 )
	{
		ALConvertCase(val, AL_UPPER, 0);
		if ( strcmp(val, XP_CFG_ON) != 0 )
			app_res.noalarm = FALSE;
		else
			app_res.noalarm = TRUE;
		return;
	}

	/*
	 * Audio command that plays audio files.
	 */
	if ( strcmp(name, XP_CFG_AUDIO) == 0 )
	{
		if ( app_res.audio_cmd )
			free(app_res.audio_cmd);
		app_res.audio_cmd = XPSafeAlloc(strlen(val)+1);
		strcpy(app_res.audio_cmd, (char *)val);
		return;
	}

	/*
	 * Calendar command configuration.
	 */
	if ( strcmp(name, XP_CFG_CAL) == 0 )
	{
		if ( app_res.calendar_cmd )
			free(app_res.calendar_cmd);
		app_res.calendar_cmd = XPSafeAlloc(strlen(val)+1);
		strcpy(app_res.calendar_cmd, (char *)val);
		return;
	}

	/*
	 * Email command configuration.
	 */
	if ( strcmp(name, XP_CFG_EMAIL) == 0 )
	{
		if ( app_res.email_cmd )
			free(app_res.email_cmd);
		app_res.email_cmd = XPSafeAlloc(strlen(val)+1);
		strcpy(app_res.email_cmd, (char *)val);
		return;
	}

	/*
	 * Email address configuration.
	 */
	if ( strcmp(name, XP_CFG_EMAIL_ADDR) == 0 )
	{
		if ( val != NULL )
		{
			if ( app_res.email_addr )
				free(app_res.email_addr);
			app_res.email_addr = XPSafeAlloc(strlen(val)+1);
			strcpy(app_res.email_addr, (char *)val);
		}
		return;
	}

	/*
	 * Printer configuration.
	 */
	if ( strcmp(name, XP_CFG_PRINT) == 0 )
	{
		if ( app_res.print_cmd )
			free( app_res.print_cmd );
		app_res.print_cmd = XPSafeAlloc(strlen(val)+1);
		strcpy(app_res.print_cmd, (char *)val);
		return;
	}

	/*
	 * Gtk rc file.
	 */
	if ( strcmp(name, XP_CFG_GTKRC) == 0 )
	{
		if ( app_res.gtk_rc )
			free(app_res.gtk_rc);
		app_res.gtk_rc = XPSafeAlloc(strlen(val)+1);
		strcpy(app_res.gtk_rc, (char *)val);
		return;
	}

}
