/* $Id$ */

/*========================================================================
 *  Copyright (c) Michael J. Hammel 1995,1996,1997,1998.
 *========================================================================
 *              FILE NAME: dialog.c
 *            DESCRIPTION: Generic Dialog handler routines
 *      DEFINED CONSTANTS: 
 *       TYPE DEFINITIONS: 
 *      MACRO DEFINITIONS: 
 *       GLOBAL VARIABLES: 
 *       PUBLIC FUNCTIONS: XPMsgWindow
 *      PRIVATE FUNCTIONS: CloseMsgWindow
 *  SOFTWARE DEPENDENCIES: 
 *  HARDWARE DEPENDENCIES: 
 *                  NOTES: 
 *
 * SPECIAL CONSIDERATIONS:
 * Set your tabstops to 3 to make the code more readable.
 *
 *========================================================================
 *
 * MODIFICATION HISTORY:
 * $Log$
 *
 *========================================================================*/
#define XP_DIALOG_C

/* === System Headers === */
#include <X11/Xlib.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <sys/stat.h>

#include <gtk/gtk.h>


/* === Project Headers === */
#include "XNotesPlus.h"
#include "debug.h"


/* === external routines === */


/* === Public routine prototypes === */


/* === Private routine prototypes === */
static void CloseMsgWindow();


/* === Global Variables === */


/* === Static Variables === */
static GtkWidget	*messagebox=NULL;
static GdkPixmap	*error_xbm;
static GdkPixmap	*info_xbm;
static GdkPixmap	*msg_xbm;
static GdkBitmap	*msg_mask, *info_mask, *error_mask;



/*========================================================================
 *	Name:			XPMsgWindow
 *	Prototype:	XPMsgWindow(char *msg, int type, 
 *							GtkWidget **return_widget, GtkWidget *parent,
 *							gpointer ok_cb, gpointer cancel_cb, gpointer help_cb)
 *
 *	Description:
 *		This routine creates and pops up a generic message window.
 *		The content of the message is passed from the caller and the
 *		type of icon to use with the message is also defined by the
 *		caller.
 *
 *	Input Arguments:
 *		char 			*msg			Message to display
 *		int 			type			Defines which icon to use in window
 *										Valid values are:
 *										XP_ERROR_TYPE    - Error message
 *										XP_MSG_TYPE	     - Generic message
 *										XP_INFO_TYPE     - Informational message
 *										XP_CONFIRM_TYPE  - Informational message
 *										XP_MSG_NOCANCEL  - Don't use cancel button
 *										XP_MSG_NOHELP    - Don't use Help button
 *										XP_MSG_NOOK      - Don't use OK button
 *		GtkWidget	parent		Parent widget, in the center of which the message
 *										window will be opened.
 *		gpointer		ok_cb			callback to use for OK button
 *		gpointer		cancel_cb	callback to use for Close button
 *		gpointer		help_cb		currently unused
 *
 *	Output Arguments:
 *		GtkWidget	**return_widget	if non-NULL, return dialog widget 
 *												to caller here
 *
 *	Return Values:
 *		None
 *
 *	Global Variables:
 *	External Routines:
 *	Method:
 *	Restrictions:
 *	Notes:
 * This routine does allow multiple dialogs to be opened.  If you 
 * call this routine you should save the dialog widget id and close
 * the dialog yourself.
 *		
 *========================================================================*/
void
XPMsgWindow(
	char			*msg,
	int			type,
	GtkWidget	**return_widget,
	GtkWidget	*parent,
	gpointer		ok_cb,
	gpointer		ok_data,
	gpointer		cancel_cb,
	gpointer		help_cb
)
{
#ifdef DEBUG
	char			fname[]="XPMsgWindow()";
#endif

	char				title[256];
	char				buf[MAXPATHLEN];
	GtkWidget		*vbox, *hbox, *accept_button, *close_button;
	GtkWidget		*label, *label_pixmap;
	GdkPixmap		*current_xbm=NULL;
	GdkBitmap		*current_mask=NULL;
	GtkStyle			*style=NULL;
	int				shell_type;
	GShell_T			gshell;


	static int		firsttime = TRUE;


	DBGEnter();

	/*
	 * Create the dialog. We'll destroy it when it is closed.
	 */
	if (type & XP_ERROR_TYPE)
		(void) sprintf ( title, "%s", XP_ERROR_TITLE );
	else
	if (type & XP_MSG_TYPE)
		(void) sprintf ( title, "%s", XP_MSG_TITLE );
	else
	if (type & XP_INFO_TYPE)
		(void) sprintf ( title, "%s", XP_INFO_TITLE );
	else
		(void) sprintf ( title, "%s", XP_CONFIRM_TITLE );

	if ( type & XP_MSG_NOCANCEL & XP_MSG_NOHELP & XP_MSG_NOOK )
	{
		shell_type = 0;
	}
	else
	{
		if ( type & XP_CONFIRM_TYPE )
		{
			shell_type =  YES_BUTTON | NO_BUTTON | HELP_BUTTON;

			if ( type & XP_MSG_NOCANCEL )
				shell_type = shell_type ^ NO_BUTTON;
	
			if ( type & XP_MSG_NOHELP )
				shell_type = shell_type ^ HELP_BUTTON;
	
			if ( type & XP_MSG_NOOK )
				shell_type = shell_type ^ YES_BUTTON;
		}
		else
		{
			shell_type =  ACCEPT_BUTTON | CLOSE_BUTTON | HELP_BUTTON;

			if ( type & XP_MSG_NOCANCEL )
				shell_type = shell_type ^ CLOSE_BUTTON;
	
			if ( type & XP_MSG_NOHELP )
				shell_type = shell_type ^ HELP_BUTTON;
	
			if ( type & XP_MSG_NOOK )
				shell_type = shell_type ^ ACCEPT_BUTTON;
		}
	}

	messagebox = XPCreateGenericShell(title, shell_type, &gshell, TRUE );
	gtk_widget_realize(messagebox);

	gtk_signal_connect (GTK_OBJECT (messagebox), "destroy",
			GTK_SIGNAL_FUNC (CloseMsgWindow), messagebox);

	vbox = gshell.vbox;
	close_button = gshell.cancel;
	accept_button = gshell.accept;

	DBGPrintf(DBG_INFO,("vbox: %d\n", (int)vbox));

	/*
	 * We want a horizontal display, so create a hbox in the vbox.
	 */
	hbox = gtk_hbox_new(FALSE, 4);
	gtk_widget_show(hbox);
	gtk_container_add(GTK_CONTAINER(vbox), hbox);

	/*
	 * First time through create the pixmaps - we need to do this after the
	 * label has been created so we can use it as the color to map the
	 * background colors to.
	 */
	if ( firsttime )
	{
		DBGPrintf(DBG_INFO,("Creating dialog pixmaps\n"));

		style = gtk_widget_get_style( messagebox );

		sprintf(buf, "%s/%s", app_res.xpm_dir, XP_MSG_XPM);
		msg_xbm = gdk_pixmap_create_from_xpm(
			messagebox->window,
			&msg_mask,
			&style->bg[GTK_STATE_NORMAL],
			buf);
		DBGPrintf(DBG_INFO,("MSG pixmap: %d\n", (int) msg_xbm));

		sprintf(buf, "%s/%s", app_res.xpm_dir, XP_INFO_XPM);
		info_xbm = gdk_pixmap_create_from_xpm(
			messagebox->window,
			&info_mask,
			&style->bg[GTK_STATE_NORMAL],
			buf);
		DBGPrintf(DBG_INFO,("INFO pixmap: %d\n", (int) info_xbm));

		sprintf(buf, "%s/%s", app_res.xpm_dir, XP_ERROR_XPM);
		error_xbm = gdk_pixmap_create_from_xpm(
			messagebox->window,
			&error_mask,
			&style->bg[GTK_STATE_NORMAL],
			buf);
		DBGPrintf(DBG_INFO,("ERROR pixmap: %d\n", (int) error_xbm));

		firsttime = FALSE;
	}

	/*
	 * Setup the buttons in the window based on the type of message
	 * window requested.
	 */
	if (( type & XP_MSG_TYPE ) || ( type & XP_CONFIRM_TYPE ))
	{
		current_xbm = msg_xbm;
		current_mask = msg_mask;
		DBGPrintf(DBG_INFO,("using MSG pixmap: %d\n", (int) current_xbm));
	}

	if ( type & XP_INFO_TYPE )
	{
		DBGPrintf(DBG_INFO,("using INFO pixmap\n"));
		current_xbm = info_xbm;
		current_mask = info_mask;
	}

	if ( type & XP_ERROR_TYPE )
	{
		DBGPrintf(DBG_INFO,("using ERROR pixmap\n"));
		current_xbm = error_xbm;
		current_mask = error_mask;
	}

	/*
	 * Set the message boxes pixmap.
	 */
	DBGPrintf(DBG_INFO,("pixmap: %d\n", (int) current_xbm));
	label_pixmap = gtk_pixmap_new(current_xbm, current_mask);
	gtk_widget_show(label_pixmap);
	gtk_container_add(GTK_CONTAINER(hbox), label_pixmap);

	sprintf(buf,"%s", msg);
	label = gtk_label_new(buf);
	gtk_widget_set_name(label, "dialog-label");
	gtk_widget_show(label);
	gtk_container_add(GTK_CONTAINER(hbox), label);

	/*
	 * Set the callbacks for the buttons.
	 */
	if ( ! (type & XP_MSG_NOCANCEL) )
	{
		if ( cancel_cb != NULL )
		{
			gtk_signal_connect (GTK_OBJECT (close_button), "clicked",
				GTK_SIGNAL_FUNC (cancel_cb), messagebox);
 
		}
		gtk_signal_connect (GTK_OBJECT (close_button), "clicked",
			GTK_SIGNAL_FUNC (CloseMsgWindow), messagebox);
	}
	if ( ! (type & XP_MSG_NOOK) )
	{
		if ( ok_cb != NULL )
		{
			gtk_signal_connect (GTK_OBJECT (accept_button), "clicked",
				GTK_SIGNAL_FUNC (ok_cb), ok_data);
		}
		gtk_signal_connect (GTK_OBJECT (accept_button), "clicked",
			GTK_SIGNAL_FUNC (CloseMsgWindow), messagebox);
	}
 

	/*
	 * Manage the message box.
	 */
	XPCenterWindow(messagebox, parent, messagebox);

	/*
	 * Pop up the dialog shell.
	 */
	gtk_widget_show(messagebox);

	/*
	 * Pass back the messagebox widgets id, if requested, so the caller
	 * can pop down the dialog.
	 */
	if ( return_widget != NULL )
		*return_widget = messagebox;
	
	DBGExit();
}


/*========================================================================
 *	Name:			CloseMsgWindow
 *	Prototype:	CloseMsgWindow()
 *
 *	Description:
 *		Default callback for the OK and Close buttons that closes the
 *		dialog window.  This gets called even if the user specifies
 *		there own callback for the buttons, so they don't have to worry
 *		about closing the window.
 *
 *	Notes:
 *		
 *========================================================================*/
static void
CloseMsgWindow(
	GtkWidget	*widget,
	gpointer		shell
)
{
	if ( GTK_WIDGET(shell) != NULL )
		gtk_widget_destroy(GTK_WIDGET(shell));
}


