/* $Id$ */

/*========================================================================
 *  Copyright (c) Michael J. Hammel 1995,1996,1997,1998.
 *========================================================================
 *              FILE NAME: getopt.c
 *            DESCRIPTION: generic parser for command line options
 *      DEFINED CONSTANTS: 
 *       TYPE DEFINITIONS: 
 *      MACRO DEFINITIONS: 
 *       GLOBAL VARIABLES: 
 *       PUBLIC FUNCTIONS: getopt_long
 *      PRIVATE FUNCTIONS: parse_traditional, parse_long
 *  SOFTWARE DEPENDENCIES: 
 *  HARDWARE DEPENDENCIES: 
 *                  NOTES: This should work just like GNU's getopt_long().
 *								 : At least I hope it does.
 *	
 *
 * SPECIAL CONSIDERATIONS:
 * Set your tabstops to 3 to make the code more readable.
 *
 *========================================================================
 *
 * MODIFICATION HISTORY:
 * $Log$
 *
 *========================================================================*/
#define GETOPT_C

/* === System Headers === */
#include <string.h>
#include <time.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>

/* === Project Headers === */
#include "getopt.h"
#include "debug.h"


/* === external routines === */


/* === Public routine prototypes === */


/* === Private routine prototypes === */
static char	parse_traditional();
static void	parse_long();


/* === Global Variables === */
int		optindex=0;
char		*optarg=NULL;
int		opterr=1;


/* === Static Variables === */



/*========================================================================
 *	Name:			getopt_long
 *	Prototype:	getopt_long(
 *						int		argc,
 *						char		**argv,
 *						char		*options,
 *						option	app_defaults[],
 *						int		*index)
 *
 *	Description:
 * Parse command line arguments in either getopt() format or using
 *	extended (--) format of GNU getopt_long().
 *
 *	Input Arguments:
 *		int		argc			number of arguments on the command line
 *		char		**argv		command line arguments
 *		char		*options		string of options that should be parsed using
 *									traditional "-x" format.
 *		option	app_defaults[]		array of struct option entries which
 *											define the long options to parse.
 *
 *	Output Arguments:
 *		int		index			index into app_defaults which the current
 *									command line option matches
 *
 *	Return Values:
 *		One of:
 *		-1			The option was not recognized as either a "-" or "--"
 *					format OR was not found in either the app_defaults 
 *					(for "--" options) or in the options string (for "-"
 *					options).
 *		0			A long option was parsed - use the index to determine
 *					which option was found.
 *		char		a single character from within the *options string which
 *					was parsed using the traditional "-x" format.
 *		EOF		no more command line arguments to parse.
 *
 *	Global Variables:
 *		int		optindex		index into argv that was last processed
 *		char		*optarg		pointer to argument associated with the current
 *									command line option being processed
 *		int		opterr		0: don't print error messages to stderr
 *									1: print error messages to stderr
 *
 *	External Routines:
 *	Method:
 *	Restrictions:
 *	Notes:
 *	This is inspired from the GNU getopt_long(), but none of the code
 * is from there.  I went looking for it on gnu.org, but couldn't find it
 * so decided to write my own version.
 *========================================================================*/
char
getopt_long(
	int		argc,
	char		**argv,
	char		*options,
	option	app_defaults[],
	int		*index
)
{
	int		chr;

	/*
	 * Increment option index.  It should now (in a perfect world)
	 * point to a command line option (and not an option argument).
	 */
	optindex++;

	/*
	 * If the index is equal to the number of arguments, then we've
	 * reached the end of the command line options.  Return an EOF.
	 * Reset the option index too.
	 */
	if (optindex == argc)
	{
		optindex = 0;
		return((char)EOF);
	}


	/*
	 * Determine the command line option format to parse.
	 * There are only three possibilities:
	 * 1. The option uses the getopt_long() "--" format.
	 * 2. The option uses the traditional "-" format.
	 * 3. The option is neither format.
	 */
	if ( strncmp(argv[optindex], "--", 2 ) == 0 )
	{
		/*
		 * Handle the getopt_long() formatted options.
		 */
		parse_long(argv, app_defaults, index);
		if ( *index == -1 )
			return((char)-1);
		else
			return((char)0);
	}
	else
	if ( strncmp(argv[optindex], "-", 1 ) == 0 )
	{
		/*
		 * Handle the traditional getopt() formatted options.
		 */
		chr = parse_traditional(argv, options);
		return((char)chr);
	}
	else
	{
		/*
		 * Unrecognized format.  If opterr is set then print a message
		 * to stderr and return a -1.
		 */
		if ( opterr )
		{
			fprintf(stderr, 
				"getopt_long(): Unknown option format (skipping): %s\n", 
				argv[optindex]);
		}
		return((char)-1);
	}
}

static char
parse_traditional(
	char		**argv,
	char		*options
)
{
	char		chr;
	char		*ptr;

	/*
	 * Grab the option character.
	 */
	bcopy((char *)(argv[optindex]+1), (char *)&chr, 1);

	/*
	 * Find it in the options string, if it exists.
	 */
	if ((ptr=strchr(options, (int)chr)) != NULL )
	{
		/*
		 * If it does exist, does it have a ":" after it?  This means
		 * that there is an options argument that goes with it.
		 */
		if ( strncmp((char *)(ptr+1), ":", 1) == 0 )
		{
			optindex++;
			optarg = argv[optindex];
		} 
		else
			optarg = NULL;
		return(chr);
	}

	/*
	 * Option was not found.
	 */
	return((char)-1);

}

static void
parse_long(
	char		**argv, 
	option	app_defaults[], 
	int		*index
)
{
	char		*name;
	int		i;

	if ( app_defaults == NULL )
	{
		*index = -1;
		return;
	}

	/*
	 * Grab the option name.
	 */
	name = (char *)(argv[optindex]+2);

	/*
	 * Find it in the app_defaults array.
	 */
	i = 0;
	while ( app_defaults[i].option_name != NULL )
	{
		if ( strcmp(app_defaults[i].option_name, name) == 0 )
		{
			/*
			 * If it does exist, does it have an option argument?
			 */
			if ( app_defaults[i].use_optarg == 1 )
			{
				optindex++;
				optarg = argv[optindex];
			} 
			else
				optarg = NULL;
			*index = i;
			return;
		}
		i++;
	}

	/*
	 * Option was not found.
	 */
	*index = -1;
	return;

}
