/* $Id$ */

/*========================================================================
 *  Copyright (c) Michael J. Hammel 1995,1996,1997,1998.
 *========================================================================
 *              FILE NAME: io.c
 *            DESCRIPTION: file i/o routines for XNotesPlus notes
 *      DEFINED CONSTANTS: 
 *       TYPE DEFINITIONS: 
 *      MACRO DEFINITIONS: 
 *       GLOBAL VARIABLES: 
 *       PUBLIC FUNCTIONS: XPOpenNote, XPExportNote
 *      PRIVATE FUNCTIONS: CreateIOPrompt, IOCancel, IOAcceptCB,
 *								 : IOWriteFile
 *  SOFTWARE DEPENDENCIES: 
 *  HARDWARE DEPENDENCIES: 
 *                  NOTES: this only handles some io functions.  notes.c
 *                  has other routines that handle io to notes files.
 *	
 *
 * SPECIAL CONSIDERATIONS:
 * Set your tabstops to 3 to make the code more readable.
 *
 *========================================================================
 *
 * MODIFICATION HISTORY:
 * $Log$
 *
 *========================================================================*/
#define XP_IO_C

/* === System Headers === */
#include <X11/Xlib.h>
#include <string.h>
#include <time.h>
#include <sys/param.h>
#ifdef USE_DIR
#include <sys/dir.h>
#else
#include <dirent.h>
#endif
#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>

/* GTK/GDK specific includes */
#include <gtk/gtk.h>
#include <gdk/gdkx.h>


/* === Project Headers === */
#include "XNotesPlus.h"
#include "debug.h"


/* === external routines === */


/* === Public routine prototypes === */
void	IOPopDown();
void	IODirCB();
void	IOFileCB();
void	IOExists();
void	IOExistsAccept();
void	IOExistsCancel();
void	IOExistsPopDown();
void	IOWriteFile();
int	FillList();


/* === Private routine prototypes === */
static void CreateIOPrompt();
static void	IOCancel();
static void	IOAcceptCB();


/* === Global Variables === */
extern AppRes		app_res;
extern StickyNote	*notes;
extern GtkWidget	*toplevel;


/* === Static Variables === */
static GtkWidget 		*io_widget = NULL;

static char			outpath[MAXPATHLEN];
static StickyNote	*localpn = (StickyNote *)0;
static int			io_type;



/*========================================================================
 *	Name:			XPOpenNote 
 *	Prototype:	XPOpenNote(GtkWidget *widget, int index)
 *					
 *	Description:
 * Open a file (ie import it) to a note
 *
 *	Input Arguments:
 *	Output Arguments:
 *	Return Values:
 *	Global Variables:
 *	External Routines:
 *	Method:
 *	Restrictions:
 *	Notes:
 *========================================================================*/
void
XPOpenNote(
	GtkWidget	*widget,
	int			index
)
{
	StickyNote	*pn;
	char			*path;
	char			freeit;

	/*
	 * If the File Selection widget is open then close it first.
	 */
	if ( io_widget != NULL )
		IOCancel();

	freeit = FALSE;

	/*
	 * Find the note we're opening.
	 */
	if ((pn = XPFindNote(index)) == NULL)
		return;

	/* get configured or current directory, whichever applies */
	if ( app_res.home_dir == NULL )
	{
		if ( getenv("HOME") == NULL )
		{
			path = "/";
		}
		else
		{
			path = (char *)XPSafeAlloc (strlen(getenv("HOME"))+1);
			strcpy(path, getenv("HOME"));
			freeit = TRUE;
		}
	}
	else
	{
		path = (char *)XPSafeAlloc (strlen(app_res.home_dir)+1);
		strcpy(path, app_res.home_dir);
		freeit = TRUE;
	}

	/*
	 * Call CreateIOPrompt() to allow user to select or name a file
	 */
	CreateIOPrompt(pn, XNOTES_OPEN_FILE, path);

	if (freeit)
		free(path);
}



/*========================================================================
 *	Name:			XPExportNote
 *	Prototype:	void XPExportNote
 *					
 *	Description:
 * Export a note to a file (ie some other file besides the notes file)
 *
 *	Input Arguments:
 *	Output Arguments:
 *	Return Values:
 *	Global Variables:
 *	External Routines:
 *	Method:
 *	Restrictions:
 *	Notes:
 *========================================================================*/
void
XPExportNote(
	GtkWidget	*widget,
	int			index
)
{
	StickyNote		*pn;
	char				*path;
	char				freeit;

	freeit = False;

	/*
	 * Find the note we're exporting.
	 */
	if ((pn = XPFindNote(index)) == NULL)
		return;

	/* get configured or current directory, whichever applies */
	if ( app_res.home_dir == NULL )
	{
		if ( getenv("HOME") == NULL )
		{
			path = "/";
		}
		else
		{
			path = (char *)XPSafeAlloc (strlen(getenv("HOME")));
			strcpy(path, getenv("HOME"));
			freeit = TRUE;
		}
	}
	else
	{
		path = (char *)XPSafeAlloc (strlen(app_res.home_dir));
		strcpy(path, app_res.home_dir);
		freeit = TRUE;
	}

	/*
	 * Call CreateIOPrompt() to allow user to select or name a file
	 */
	CreateIOPrompt(pn, XNOTES_EXPORT_FILE, path);

	if (freeit)
		free(path);
}



/*========================================================================
 *	Name:			CreateIOPrompt
 *	Prototype:	CreateIOPrompt(StickyNote *pn, int type, char *path)
 *					
 *	Description:
 * Create a file selection widget
 *
 *	Input Arguments:
 * StickyNote	*pn		pointer to a StickyNote structure
 * int			type		either XNOTES_OPEN_FILE, or XNOTES_EXPORT_FILE
 * char			*path		path to directory which is to be used in file selection
 *
 *	Output Arguments:
 *	Return Values:
 *	Global Variables:
 *	External Routines:
 *	Method:
 *	Restrictions:
 *	Notes:
 *========================================================================*/
static void
CreateIOPrompt(
	StickyNote 	*pn,
	int			type,
	char			*path
)
{

	/*
	 * Save the type in static storage for the callbacks to use 
	 */
	io_type = type;
	localpn = pn;

	/* get mouse location */
	localpn = pn;

	/*
	 * Create a popup shell widget
	 */
	io_widget = gtk_file_selection_new (NotesIOPrompt);

	/*
	 * Initialize the file selection window based on the path
	 * passed from the caller.
	 */
	gtk_file_selection_set_filename (GTK_FILE_SELECTION(io_widget), path);

	/* The Cancel/Accept buttons */
	gtk_signal_connect_object (
		GTK_OBJECT (GTK_FILE_SELECTION (io_widget)->cancel_button), "clicked", 
		(GtkSignalFunc) IOCancel, NULL);
	gtk_signal_connect_object (
		GTK_OBJECT (GTK_FILE_SELECTION (io_widget)->ok_button), "clicked", 
		(GtkSignalFunc) IOAcceptCB, NULL);

	/*
	 * Pop up the window.  The File Seletion widget doesn't respond to
	 * window positioning requests, unfortunately.
	 */
	gtk_widget_show( io_widget );
 
}


/*========================================================================
 *	Name:			IOCancel
 *	Prototype:	IOCancel(GtkWidget *widget, gpointer data)
 *					
 *	Description:
 * Close the File Selection Window
 *
 *========================================================================*/
static void
IOCancel(
	GtkWidget	*widget,
	gpointer		data
)
{
	if ( io_widget != NULL )
	{
		gtk_widget_destroy(io_widget);
		io_widget = NULL;
	}
}



/*========================================================================
 *	Name:			IOAcceptCB
 *	Prototype:	IOAcceptCB(GtkWidget *widget, gpointer data)
 *					
 *	Description:
 * Process the file selected
 *
 *========================================================================*/
static void
IOAcceptCB(
	GtkWidget	*widget,
	gpointer		data
)
{
#ifdef DEBUG
	char	*fname="IOAcceptCB()";
#endif

	StickyNote 	*pn;
	char			path[MAXPATHLEN];
	char			output[128+MAXPATHLEN];
	struct stat	stat_buf;
	char			*buf;

	DBGEnter();

	/*
	 * Grab saved pointer to note structure.
	 */
	pn = localpn;

	/* get the path name */
	buf = gtk_file_selection_get_filename (GTK_FILE_SELECTION(io_widget));

	switch (io_type)
	{
		case XNOTES_OPEN_FILE:
			DBGPrintf(DBG_INFO,("Importing a file: %s\n", buf));

			if ( strlen(buf) > MAXPATHLEN )
			{
				sprintf(output, "File name (%s) is too long!", buf);
				XPErrPopUp(output);
				return;
			}
			else
				strcpy ( path, buf );

			/*
			 * make sure file exists one more time
			 * and check if its a file or directory.
			 */
			if ( stat(path, &stat_buf) == 0 )
			{
				if ( stat_buf.st_mode & S_IFDIR )
				{
					/* just reset the file-selection window */
					gtk_file_selection_set_filename (
						GTK_FILE_SELECTION(io_widget), path);
				}
				else
				{
					/*
					 * Insert the text of the file selected into the note
					 * and close the File Selection window.
					 *
					 * InsertText is a generic routine, called from various
					 * places, so we have to handle closing the File Selection
					 * window ourselves.
					 */
					XPInsertText( pn, path, FALSE, FILE_TYPE );
					gtk_widget_set_sensitive ( pn->pn_savewidget, TRUE );
					IOCancel(NULL, NULL);
				}
			}
			else
			{
				sprintf(output, " %s does not exist! ", path);
				XPErrPopUp(output);
			}
			break;

		case XNOTES_EXPORT_FILE:

			DBGPrintf(DBG_INFO,("Exporting a file: %s\n", buf));

			if ( strlen(buf) > MAXPATHLEN )
			{
				sprintf(output, "File name (%s) is too long!", buf);
				XPErrPopUp(output);
				return;
			}
			else
				strcpy ( path, buf );

			/*
			 * check if file exists and if its a directory.
			 */
			DBGPrintf(DBG_INFO,("path: %s\n", path));
			if ( stat(path, &stat_buf) == 0 )
			{
				if ( stat_buf.st_mode & S_IFDIR )
				{
					/* just reset the file-selection window */
					gtk_file_selection_set_filename (
						GTK_FILE_SELECTION(io_widget), path);
					return;
				}
				else
				{
					/*
					 * File exists - ask the user if they want to over write
					 * it.
					 */
					strcpy(outpath, path);
					sprintf(output, "%s: %s", XP_S_FILE_EXIST, path);
					XPMsgWindow(
						output, XP_MSG_TYPE | XP_MSG_NOHELP, NULL,
						pn->pn_shellwidget, IOWriteFile, NULL, NULL, NULL);
					return;
				}
			}

			/*
			 * File does not exist - go ahead and write to that file.
			 */
			strcpy(outpath, path);
			IOWriteFile();

			break;
	}

}


/*========================================================================
 *	Name:			IOWriteFile
 *	Prototype:	IOWriteFile()
 *					
 *	Description:
 * Write the note to a file
 *
 *========================================================================*/
void
IOWriteFile()
{
#ifdef DEBUG
	char	*fname="IOWriteFile()";
#endif

	StickyNote	*pn;
	FILE			*fp;
	int			len;
	char			path[MAXPATHLEN];
	char			output[128+MAXPATHLEN];

	DBGEnter();

	/*
	 * Save local copies of the note and path of interest.
	 */
	strcpy (path, outpath);
	pn = localpn;

	/*
	 * Get the length of the text in the window.
	 */
	if (pn->pn_text != NULL )
		free(pn->pn_text);
	pn->pn_text = gtk_editable_get_chars(GTK_EDITABLE(pn->pn_textwidget), 0, -1);
	len = strlen(pn->pn_text);

	/*
	 * Open the first temporary file for writing
	 */
	if ((fp = fopen(path, "w")) == NULL) 
	{
		sprintf(output, "Can't open %s for writing!", path);
		XPErrPopUp(output);
		fclose(fp);
		return;
	}

	/*
	 * Write out the text of the note.
	 */
	if (len)
		fwrite(pn->pn_text, sizeof(char), len, fp);

	/* add a newline and terminator */
	fwrite("\n", sizeof(char), strlen("\n"), fp);

	fclose(fp);

	/*
	 * Close the file selection dialog, if its open.
	 */
	IOCancel(NULL, NULL);
}
