/* $Id$ */

/*========================================================================
 *  Copyright (c) Michael J. Hammel 1995,1996,1997,1998.
 *========================================================================
 *              FILE NAME: colors.c
 *            DESCRIPTION: functions that handle application colors 
 *      DEFINED CONSTANTS: 
 *       TYPE DEFINITIONS: 
 *      MACRO DEFINITIONS: 
 *       GLOBAL VARIABLES: 
 *       PUBLIC FUNCTIONS: XPSetGlobalColors
 *      PRIVATE FUNCTIONS: 
 *  SOFTWARE DEPENDENCIES: X11R4 or later, Xt, Motif 1.2
 *  HARDWARE DEPENDENCIES: 
 *                  NOTES: set tabstops=3 for readibility
 *
 * SPECIAL CONSIDERATIONS:
 * Set your tabstops to 3 to make the code more readable.
 *========================================================================
 *
 * MODIFICATION HISTORY:
 *	$Log$
 *
 *========================================================================*/
#ifndef XP_COLORS_C
#define XP_COLORS_C

/* === System Headers === */
#include <stdio.h>
#include <stdlib.h>

#include "motif.h"

/* === Project Headers === */
#include "XNotesPlus.h"
#include "debug.h"

/* === external routines === */

/* === Public routine prototypes === */

/* === Private routine prototypes === */

/* === Global Variables defined externally === */
extern Widget			Top_Level;		/* shell of main window */
extern AppRes			app_res;			/* application resource data */

/* === Global Variables defined locally === */

/* === Static Variables === */


/*========================================================================
 *	Name:			XPSetGlobalColors
 *	Prototype:	int XPSetGlobalColors()
 *
 *	Description:
 *		Establish the colormap we'll be using for all windows.
 *
 *	Input Arguments:
 *		Widget		app_widget		Widget with colormap to copy from.
 *		Colormap		*caller_cmap	where to place the new colormap
 *		int			max_colors		number of colors to reserver for our use.
 *
 *	Output Arguments:
 *	Return Values:
 *		XP_VISUAL_NOT_SUPPORTED		Default visual is not supported.
 *		XP_CMAP_SMALL_IMMUTABLE		Colormap can't be changed and only
 *											has 256 indices available.
 *		XP_CMAP_BIG_IMMUTABLE		Colormap can't be changed but
 *											has > 256 indices available.
 *		XP_CMAP_OK						Default colormap should be fine.
 *		XP_E_SUCCESS					Installed new colormap.
 *		
 *	Global Variables:
 *	External Routines:
 *	Method:
 *	Restrictions:
 *	Notes:
 *		Assumes that if it is called, the caller requires a private 
 *		colormap to be installed.
 *		
 *========================================================================*/
int
XPSetGlobalColors(
	Widget	app_widget,
	Colormap	*caller_cmap,
	int		max_colors
)
{
#ifdef DEBUG
	char			fname[]="XPSetGlobalColors()";	/* name of this routine */
#endif


	Visual		*default_visual;
	Display		*dpy;
	Window		rwin;
	int			screen;
	XColor		palette[256];
	XColor		default_palette[256];
	Colormap		cmap;
	int			i;

	DBGEnter();

	/*
	 * Fetch things we will need often, like the display and RootWindow.
	 */
	dpy  = XtDisplay(app_widget);
	rwin = DefaultRootWindow(dpy);
	screen = DefaultScreen(dpy);

	/*
	 * Get the default visual info.
	 */
	default_visual = DefaultVisual(dpy, screen);

	/*
	 * Test the number of color planes there are.
	 * We don't support anything less than 8 color planes
	 * (ie 256 colors or more only).
	 */
	if ( DefaultDepth(dpy, screen) < 8 )
		return(XP_VISUAL_NOT_SUPPORTED);

	/*
	 * Test the visual type
	 */
	switch(default_visual->class)
	{
		/*
		 * The first three types have colormaps that can't be changed.
		 * The first two will cause color problems most likely.
		 * TrueColor may cause problems, but only if the number of color
		 * planes is less than 8.  Most TrueColor visuals have 24
		 * or more planes.
		 */
		case StaticGray:
		case StaticColor:
			DBGPrintf(DBG_TRACE1, ("Visual type: Static (Gray or Color)\n"));
			return(XP_CMAP_SMALL_IMMUTABLE);
			break;
		case TrueColor:
			if ( DefaultDepth(dpy, screen) < 8 )
			{
				DBGPrintf(DBG_TRACE1, ("Visual type: small True Color\n"));
				return(XP_CMAP_SMALL_IMMUTABLE);
			}
			else
			{
				DBGPrintf(DBG_TRACE1, ("Visual type: big True Color\n"));
				return(XP_CMAP_BIG_IMMUTABLE);
			}
			break;

		/*
		 * The next three can be modified.  The first two are limited
		 * color indices, usually around 256 colors total.  The DirectColor
		 * is the one we dream about - millions of colors (potentially)
		 * and its modifiable!  We pretty much can't run out of colors in
		 * that case.
		 */
		case GrayScale:
		case PseudoColor:
			DBGPrintf(DBG_TRACE1, ("Visual type: Indexed\n"));
			break;
			
		case DirectColor:
			DBGPrintf(DBG_TRACE1, ("Visual type: Direct\n"));
			return(XP_CMAP_OK);
			break;
	}

	/*
	 * Should we get here, we know we have either a PseudoColor or a 
	 * GrayScale display.  We need to allocate a colormap, copy the
	 * first 256-max_colors entries from the default colormap to it, then
	 * apply it to the top level window so all child windows will inherit
	 * it.  
 	 *
	 * Well, thats the theory anyway.
	 */

	/*
	 * Create a new colormap.  If we neeed a private colormap we
	 * allocate all the colorcells.  Otherwise we don't allocate any
	 * cells.
	 */
	cmap = 
		XCopyColormapAndFree(dpy, 
			DefaultColormapOfScreen (XtScreen(Top_Level)) );

	/*
	 * Initialize the pixel ids for the default color array.
	 */
	for (i=0; i<256; i++)
		default_palette[i].pixel = i;

	/*
	 * Grab a copy of the current colormap colors.
	 */
	XQueryColors(dpy, DefaultColormap(dpy, screen), default_palette,
			DisplayCells(dpy, screen));
	
	/*
	 * Use the copy to store up to 256 - max_colors of the default colors 
	 * into the new colormap.  This leaves us with up to max_colors colors 
	 * to allocate for our own use.
	 */
	for (i=0; i<(DisplayCells(dpy,screen)-max_colors); i++)
	{
		palette[i].pixel = default_palette[i].pixel;
		palette[i].red   = default_palette[i].red;
		palette[i].green = default_palette[i].green;
		palette[i].blue  = default_palette[i].blue;
		palette[i].flags = default_palette[i].flags;
		XAllocColor(dpy, cmap, &(palette[i]));
	}

	/*
	 * Save cmap value.
	 */
	*caller_cmap = cmap;

	return(XP_E_SUCCESS);
}

#endif /* XP_COLORS_C */
