/* $Id$ */

/*========================================================================
 *  Copyright (c) Michael J. Hammel 1995,1996,1997,1998.
 *========================================================================
 *              FILE NAME: io.c
 *            DESCRIPTION: file i/o routines for XNotesPlus notes
 *      DEFINED CONSTANTS: 
 *       TYPE DEFINITIONS: 
 *      MACRO DEFINITIONS: 
 *       GLOBAL VARIABLES: 
 *       PUBLIC FUNCTIONS: XPOpenNote, XPExportNote
 *      PRIVATE FUNCTIONS: CreateIOPrompt, IOCancel, IOAcceptCB,
 *								 : IOWriteFile
 *  SOFTWARE DEPENDENCIES: 
 *  HARDWARE DEPENDENCIES: 
 *                  NOTES: this only handles some io functions.  notes.c
 *                  has other routines that handle io to notes files.
 *	
 *
 * SPECIAL CONSIDERATIONS:
 * Set your tabstops to 3 to make the code more readable.
 *
 *========================================================================
 *
 * MODIFICATION HISTORY:
 * $Log$
 *
 *========================================================================*/
#define XP_IO_C

/* === System Headers === */
#include <X11/Xlib.h>
#include <string.h>
#include <time.h>
#include <sys/param.h>
#ifdef USE_DIR
#include <sys/dir.h>
#else
#include <dirent.h>
#endif
#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>

#include "motif.h"


/* === Project Headers === */
#include "XNotesPlus.h"
#include "debug.h"


/* === external routines === */


/* === Public routine prototypes === */
void	IOPopDown();
void	IODirCB();
void	IOFileCB();
void	IOExists();
void	IOExistsAccept();
void	IOExistsCancel();
void	IOExistsPopDown();
void	IOWriteFile();
int	FillList();


/* === Private routine prototypes === */
static void CreateIOPrompt();
static void	IOCancel();
static void	IOAcceptCB();


/* === Global Variables === */
extern AppRes		app_res;
extern StickyNote	*notes;
extern Widget		Top_Level;


/* === Static Variables === */
static Widget 		io_widget = NULL;

static char			outpath[MAXPATHLEN];
static StickyNote	*localpn = (StickyNote *)0;
static int			io_type;



/*========================================================================
 *	Name:			XPOpenNote 
 *	Prototype:	XPOpenNote(Widget widget, int index, XtPointer call_data)
 *					
 *	Description:
 * Open a file (ie import it) to a note
 *
 *	Input Arguments:
 *	Output Arguments:
 *	Return Values:
 *	Global Variables:
 *	External Routines:
 *	Method:
 *	Restrictions:
 *	Notes:
 *========================================================================*/
void
XPOpenNote(
	Widget		widget,
	int			index, 
	XtPointer	call_data
)
{
#ifdef DEBUG
	char		*fname="XPOpenNote()";
#endif

	StickyNote	*pn;
	char			*path;
	char			freeit;

	DBGEnter();

	/*
	 * If the File Selection widget is open then close it first.
	 */
	if ( io_widget != NULL )
		IOCancel();

	freeit = FALSE;

	/*
	 * Find the note we're opening.
	 */
	if ((pn = XPFindNote(index)) == NULL)
		return;

	/* get configured or current directory, whichever applies */
	if ( app_res.home_dir == NULL )
	{
		if ( getenv("HOME") == NULL )
		{
			path = "/";
		}
		else
		{
			path = (char *)XPSafeAlloc (strlen(getenv("HOME"))+1);
			strcpy(path, getenv("HOME"));
			freeit = TRUE;
		}
	}
	else
	{
		path = (char *)XPSafeAlloc (strlen(app_res.home_dir)+1);
		strcpy(path, app_res.home_dir);
		freeit = TRUE;
	}

	/*
	 * Call CreateIOPrompt() to allow user to select or name a file
	 */
	DBGPrintf(DBG_INFO,("Calling CreateIOPrompt() for XNOTES_OPEN_FILE\n"));
	CreateIOPrompt(pn, XNOTES_OPEN_FILE, path);

	if (freeit)
		free(path);
}



/*========================================================================
 *	Name:			XPExportNote
 *	Prototype:	void XPExportNote
 *					
 *	Description:
 * Export a note to a file (ie some other file besides the notes file)
 *
 *	Input Arguments:
 *	Output Arguments:
 *	Return Values:
 *	Global Variables:
 *	External Routines:
 *	Method:
 *	Restrictions:
 *	Notes:
 *========================================================================*/
void
XPExportNote(
	Widget		widget,
	int			index,
	XtPointer	call_data
)
{
#ifdef DEBUG
	char		*fname="XPExportNote()";
#endif

	StickyNote		*pn;
	char				*path;
	char				freeit;

	freeit = False;

	/*
	 * Find the note we're exporting.
	 */
	if ((pn = XPFindNote(index)) == NULL)
		return;

	/* get configured or current directory, whichever applies */
	if ( app_res.home_dir == NULL )
	{
		if ( getenv("HOME") == NULL )
		{
			path = "/";
		}
		else
		{
			path = (char *)XPSafeAlloc (strlen(getenv("HOME")));
			strcpy(path, getenv("HOME"));
			freeit = TRUE;
		}
	}
	else
	{
		path = (char *)XPSafeAlloc (strlen(app_res.home_dir));
		strcpy(path, app_res.home_dir);
			freeit = TRUE;
	}

	/*
	 * Call CreateIOPrompt() to allow user to select or name a file
	 */
	DBGPrintf(DBG_INFO,("Calling CreateIOPrompt() for XNOTES_EXPORT_FILE\n"));
	CreateIOPrompt(pn, XNOTES_EXPORT_FILE, path);

	if (freeit)
		free(path);
}



/*========================================================================
 *	Name:			CreateIOPrompt
 *	Prototype:	CreateIOPrompt(StickyNote *pn, int type, char *path)
 *					
 *	Description:
 * Create a file selection widget
 *
 *	Input Arguments:
 * StickyNote	*pn		pointer to a StickyNote structure
 * int			type		either XNOTES_OPEN_FILE, or XNOTES_EXPORT_FILE
 * char			*path		path to directory which is to be used in file selection
 *
 *	Output Arguments:
 *	Return Values:
 *	Global Variables:
 *	External Routines:
 *	Method:
 *	Restrictions:
 *	Notes:
 *========================================================================*/
static void
CreateIOPrompt(
	StickyNote 	*pn,
	int			type,
	char			*path
)
{
	Arg			args[2];
	int			nargs;
	Widget		help;
	XmString		xm_string;

	/*
	 * Only one copy at a time allowed.
	 */
	if ( io_widget != NULL )
		return;

	/*
	 * Save the type in static storage for the callbacks to use 
	 */
	io_type = type;
	localpn = pn;

	/*
	 * Create File Selection Box.
	 */
	xm_string = XmStringCreateSimple(path);
	nargs = 0;
	XtSetArg ( args[nargs], XmNdirectory, xm_string); nargs++;
	XtSetArg ( args[nargs], XmNtitle, NotesIOPrompt); nargs++;
	io_widget = XmCreateFileSelectionDialog(
						Top_Level,
						"FSWin",
						args, nargs);
	XmStringFree(xm_string);
	
	/*
	 * Get the button widgets from the dialog.
	 * Hide the Help button - we don't need it here.
	 */
	help = XmFileSelectionBoxGetChild(io_widget, XmDIALOG_HELP_BUTTON);
	XtUnmanageChild(help);

	/* The Cancel/Accept buttons */
	XtAddCallback( io_widget, XmNcancelCallback, IOCancel, (XtPointer)NULL);
	XtAddCallback( io_widget, XmNokCallback, IOAcceptCB, (XtPointer)NULL);

	/*
	 * pop up the window.
	 */
	XtManageChild(io_widget);
	XPCenterWindow(io_widget,NULL, XtParent(io_widget));
}


/*========================================================================
 *	Name:			IOCancel
 *	Prototype:	IOCancel(Widget widget, XtPointer client_data, XtPointer call_data)
 *					
 *	Description:
 * Close the File Selection Window
 *
 *========================================================================*/
static void
IOCancel(
	Widget		widget,
	XtPointer	client_data,
	XtPointer	call_data
)
{
	if ( io_widget != NULL )
	{
		XtDestroyWidget(io_widget);
		io_widget = NULL;
	}
}



/*========================================================================
 *	Name:			IOAcceptCB
 *	Prototype:	IOAcceptCB(Widget widget, XtPointer client_data, 
 *							XtPointer call_data)
 *					
 *	Description:
 * Process the file selected
 *
 *========================================================================*/
static void
IOAcceptCB(
	Widget			widget,
	XtPointer		client_data,
	XtPointer		call_data
)
{
#ifdef DEBUG
	char		*fname="IOAcceptCB()";
#endif

	StickyNote 	*pn;
	char			path[MAXPATHLEN];
	char			output[128+MAXPATHLEN];
	struct stat	stat_buf;
	char			*buf;
	XmString		xm_string;

	XmFileSelectionBoxCallbackStruct *cbs =
		(XmFileSelectionBoxCallbackStruct *)call_data;

	DBGEnter();

	/*
	 * Grab saved pointer to note structure.
	 */
	pn = localpn;

	/* get the path name */
	if ( !XmStringGetLtoR(cbs->value, XmFONTLIST_DEFAULT_TAG, &buf) )
		return;

	DBGPrintf(DBG_INFO,("io_type: %d\n", io_type));
	switch (io_type)
	{
		case XNOTES_OPEN_FILE:
			DBGPrintf(DBG_INFO,("Opening file...\n"));
			if ( strlen(buf) > MAXPATHLEN )
			{
				sprintf(output, "File name (%s) is too long!", buf);
				XPErrPopUp(output);
				return;
			}
			else
				strcpy ( path, buf );
			DBGPrintf(DBG_INFO,("path: %s\n", path));

			/*
			 * make sure file exists one more time
			 * and check if its a file or directory.
			 */
			if ( stat(path, &stat_buf) == 0 )
			{
				if ( stat_buf.st_mode & S_IFDIR )
				{
					DBGPrintf(DBG_INFO,("resetting fswindow: %s\n", path));

					/* just reset the file-selection window */
					xm_string = XmStringCreateSimple(path);
					XmFileSelectionDoSearch(io_widget, xm_string);
					XmStringFree(xm_string);
					return;
				}
				else
				{
					DBGPrintf(DBG_INFO,("inserting text\n"));
					/*
					 * Insert the text of the file selected into the note
					 * and close the File Selection window.
					 *
					 * InsertText is a generic routine, called from various
					 * places, so we have to handle closing the File Selection
					 * window ourselves.
					 */
					XPInsertText( pn, path, FALSE, FILE_TYPE );
					IOCancel(NULL, NULL, NULL);
				}
			}
			else
			{
				DBGPrintf(DBG_INFO,("file does not exist\n"));
				sprintf(output, " %s does not exist! ", path);
				XPErrPopUp(output);
			}
			DBGPrintf(DBG_INFO,("done with XNOTES_OPEN_FILE case.\n"));
			break;

		case XNOTES_EXPORT_FILE:
			DBGPrintf(DBG_INFO,("Exporting file...\n"));
			if ( strlen(buf) > MAXPATHLEN )
			{
				sprintf(output, "File name (%s) is too long!", buf);
				XPErrPopUp(output);
				return;
			}
			else
				strcpy ( path, buf );

			/*
			 * check if file exists and if its a directory.
			 */
			if ( stat(path, &stat_buf) == 0 )
			{
				if ( stat_buf.st_mode & S_IFDIR )
				{
					/* just reset the file-selection window */
					xm_string = XmStringCreateSimple(path);
					XmFileSelectionDoSearch(io_widget, xm_string);
					XmStringFree(xm_string);
					return;
				}
				else
				{
					/*
					 * File exists - ask the user if they want to over write
					 * it.
					 */
					strcpy(outpath, path);
					sprintf(output, "%s: %s", XP_S_FILE_EXIST, path);
					XPMsgWindow(
						output, XP_MSG_TYPE | XP_MSG_NOHELP, NULL, 
						pn->pn_shellwidget, IOWriteFile, NULL, NULL, NULL);
					return;
				}
			}

			/*
			 * File does not exist - go ahead and write to that file.
			 */
			strcpy(outpath, path);
			IOWriteFile();

			break;
	}

}


/*========================================================================
 *	Name:			IOWriteFile
 *	Prototype:	IOWriteFile()
 *					
 *	Description:
 * Write the note to a file
 *
 *========================================================================*/
void
IOWriteFile()
{
	StickyNote	*pn;
	FILE			*fp;
	int			len;
	char			path[MAXPATHLEN];
	char			output[128+MAXPATHLEN];

	/*
	 * Save local copies of the note and path of interest.
	 */
	strcpy (path, outpath);
	pn = localpn;

	/*
	 * Get the length of the text in the window.
	 */
	if ( pn->pn_text != NULL )
		free(pn->pn_text);
	pn->pn_text = XmTextGetString(pn->pn_textwidget);
	len = strlen(pn->pn_text);

	/*
	 * Open the first temporary file for writing
	 */
	if ((fp = fopen(path, "w")) == NULL) 
	{
		sprintf(output, "Can't open %s for writing!", path);
		XPErrPopUp(output);
		fclose(fp);
		return;
	}

	/*
	 * Write out the text of the note.
	 */
	if (len)
		fwrite(pn->pn_text, sizeof(char), len, fp);

	/* add a newline and terminator */
	fwrite("\n", sizeof(char), strlen("\n"), fp);

	fclose(fp);

	/*
	 * Close the file selection dialog, if its open.
	 */
	IOCancel(NULL, NULL, NULL);
}
